#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

require 'set'

# 変数に重複して割り当てないよう、必要な条件を列挙する。
# 例えば、enum_duplication(["A", "B", "C"], [1, 2, 4, 5, 6, 7]) とすると
# （使ってよい残る数は3, 8, 9）
# - 「Aは3でないか、Bは3でない」
# - 「Aは3でないか、Cは3でない」
# - 「Bは3でないか、Cは3でない」
# - 「Aは8でないか、Bは8でない」
# - 「Aは8でないか、Cは8でない」
# - 「Bは8でないか、Cは8でない」
# - 「Aは9でないか、Bは9でない」
# - 「Aは9でないか、Cは9でない」
# - 「Bは9でないか、Cは9でない」
# の9つの条件を返す。
def enum_duplication(empty_cells, used_numbers)
  remained_numbers = Set.new(1..9) - used_numbers
  if remained_numbers.size != empty_cells.size
    raise "Invalid number of candidates"
  end
  
  result = []
  remained_numbers.each do |n|
    empty_cells.combination(2).each do |a, b|
      result << {:number => n, :variables => [a, b]}
    end
  end
  result
end

# 問題ファイルを読み込む
problem = ARGF.readlines.map{ |line| line.chomp }
while problem.last.empty?
  problem.pop
end

# 問題の大きさが適切か検査する
unless problem.size == 9 && problem.all?{ |line| line =~ /\A[.0-9]{9}\z/ }
  raise "Invalid problem format"
end

# 数値での表現に変換する
problem.map!{ |line| line.each_char.map{ |c| c == '.' ? nil : c.to_i } }

# 各マスについて、入りうる数を列挙する(1)一マス単位での検査
candidates = []

for i in 0..8
  candidates[i] = []
  for j in 0..8
    # すでに数の入っているマスの場合
    next if problem[i][j]
    
    candidates[i][j] = Set.new(1..9)
    
    # 同じ行ですでに使われている数は使えない
    candidates[i][j] -= problem[i]
    
    # 同じ列ですでに使われている数は使えない
    candidates[i][j] -= problem.map{ |row| row[j] }
    
    # 同じブロックですでに使われている数は使えない
    block_row_begin = (i/3) * 3
    block_col_begin = (j/3) * 3
    for r in block_row_begin..(block_row_begin+2)
      candidates[i][j] -= problem[r][block_col_begin..(block_col_begin+2)]
    end
    
    if candidates[i][j].empty?
      raise "No candidate at (#{i}, #{j})"
    end
  end
end

# 各マスについて、入りうる数を列挙する(2)行、列、ブロック単位での検査
# （一つの行内に複数の空きマスがある場合、それらに同じ数が入らないようにする）
duplication = []

for i in 0..8
  empty_cells = []
  used_numbers = []
  for j in 0..8
    if problem[i][j]
      used_numbers << problem[i][j]
    else
      empty_cells << [i, j]
    end
  end
  duplication.concat enum_duplication(empty_cells, used_numbers)
end

for j in 0..8
  empty_cells = []
  used_numbers = []
  for i in 0..8
    if problem[i][j]
      used_numbers << problem[i][j]
    else
      empty_cells << [i, j]
    end
  end
  duplication.concat enum_duplication(empty_cells, used_numbers)
end

for r in 0..2
  for c in 0..2
    empty_cells = []
    used_numbers = []
    
    for y in 0..2
      for x in 0..2
        i = r*3+y
        j = c*3+x
        
        if problem[i][j]
          used_numbers << problem[i][j]
        else
          empty_cells << [i, j]
        end
      end
    end
    duplication.concat enum_duplication(empty_cells, used_numbers)
  end
end

# 連言標準形にする
# cnf = [["foo", "-bar"], ["-foo", "buz"]]
# は、 [foo or not(bar)] and [not(foo) and buz] の意味
cnf = []

# (1)の制約
for i in 0..8
  for j in 0..8
    next unless candidates[i][j]
    
    # 候補のうち少なくとも1つは選ばれる
    cnf << candidates[i][j].map{ |d| "CELL_#{i}#{j}_IS_#{d}" }
    
    # 候補が2つ選ばれることはない
    candidates[i][j].to_a.combination(2) do |f|
      cnf << f.map{ |d| "-CELL_#{i}#{j}_IS_#{d}" }
    end
    
    # 候補でないものは選ばれない
    (Set.new(1..9) - candidates[i][j]).each do |d|
      cnf << ["-CELL_#{i}#{j}_IS_#{d}"]
    end
  end
end

# (2)の制約
duplication.each do |dupcase|
  cnf << dupcase[:variables].map{ |d| "-CELL_#{d[0]}#{d[1]}_IS_#{dupcase[:number]}" }
end

# dimacs形式で出力する。
# dimacs形式では、変数はすべて自然数に直しておかないとならないので、
# それをvariable_tableとして保存している。
variable_array = cnf.flatten.uniq.map{ |v| v[0] == '-' ? v[1..-1] : v }.uniq.sort
variable_table = {}
variable_array.each_with_index do |v, i|
  variable_table[v] = i + 1
end

dimacs_main = ""

cnf.each do |f|
  f.each do |v|
    if v[0] == '-'
      name = v[1..-1]
      sign = '-'
    else
      name = v
      sign = ''
    end
    
    dimacs_main << "#{sign}#{variable_table[name]} "
  end
  dimacs_main << "0\n"
end

puts "c ---------- PROBLEM ----------"
puts "c   | 0 1 2 3 4 5 6 7 8"
puts "c --|-------------------"
for i in 0..8
  puts "c #{i} | #{problem[i].map{ |c| c || '.' }.join(' ')}"
end
puts "c ---------- VARIABLES ----------"
variable_table.each_pair do |variable, num|
  puts "c #{num} = #{variable}"
end
puts "c ---------- CONSTRAINT ----------"
puts "p cnf #{variable_table.size} #{cnf.size}"
puts dimacs_main

